@@ -1,16 +1,73 @@
+use bitflags::bitflags;
 #[cfg(feature = "embedded")]
 use core::convert::Infallible;
-#[cfg(feature = "embedded")]
-use embedded_hal::serial;
-#[cfg(feature = "embedded")]
-use nb;
-
-#[cfg(feature = "fmt")]
-use core::fmt;
+use core::fmt::{self, Display, Formatter};
 
 use crate::registers::Registers;
 
-#[derive(Debug, Clone, PartialEq)]
+bitflags! {
+    /// Interrupt Enable Register (bitflags)
+    pub struct IER: u8 {
+        /// Enable Received Data Available Interrupt
+        const RDAI  = 0b0000_0001;
+        /// Enable Transmitter Holding Register Empty Interrupt
+        const THREI = 0b0000_0010;
+        /// Enable Receiver Line Status Interrupt
+        const RLSI  = 0b0000_0100;
+        /// Enable Modem Status Interrupt
+        const MSI   = 0b0000_1000;
+        /// Enable Sleep Mode (16750)
+        const SM    = 0b0001_0000;
+        /// Enable Low Power Mode (16750)
+        const LPM   = 0b0010_0000;
+    }
+}
+
+bitflags! {
+    /// Line Status Register (bitflags)
+    pub struct LSR: u8 {
+        /// Data Ready
+        const DR = 0b0000_0001;
+        /// Overrun Error
+        const OE = 0b0000_0010;
+        /// Parity Error
+        const PE = 0b0000_0100;
+        /// Framing Error
+        const FE = 0b0000_1000;
+        /// Break Interrupt
+        const BI = 0b0001_0000;
+        /// Transmitter Holding Register Empty
+        const THRE = 0b0010_0000;
+        /// Data Holding Regiters Empty
+        const DHRE = 0b0100_0000;
+        /// Error in Received FIFO
+        const RFE = 0b1000_0000;
+    }
+}
+
+bitflags! {
+    /// Modem Status Register (bitflags)
+    pub struct MSR: u8 {
+        /// Delta Clear To Send
+        const DCTS = 0b0000_0001;
+        ///Delta Data Set Ready
+        const DDSR = 0b0000_0010;
+        ///Trailing Edge Ring Indicator
+        const TERI = 0b0000_0100;
+        ///Delta Data Carrier Detect
+        const DDCD = 0b0000_1000;
+        ///Clear To Send
+        const CTS = 0b0001_0000;
+        ///Data Set Ready
+        const DSR = 0b0010_0000;
+        ///Ring Indicator
+        const RI = 0b0100_0000;
+        ///Carrier Detect
+        const CD = 0b1000_0000;
+    }
+}
+
+#[derive(Copy, Clone, Debug, Eq, PartialEq)]
 pub enum ChipFifoInfo {
     NoFifo,
     Reserved,
@@ -18,7 +75,7 @@ pub enum ChipFifoInfo {
     Enabled,
 }
 
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Copy, Clone, Debug, Eq, PartialEq)]
 pub enum InterruptType {
     ModemStatus,
     TransmitterHoldingRegisterEmpty,
@@ -28,7 +85,7 @@ pub enum InterruptType {
     Reserved,
 }
 
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Copy, Clone, Debug, Eq, PartialEq)]
 pub enum Parity {
     No,
     Odd,
@@ -37,30 +94,48 @@ pub enum Parity {
     Space,
 }
 
-/// # MMIO version of uart 8250
+/// An error encountered which trying to transmit data.
+#[derive(Copy, Clone, Debug, Eq, PartialEq)]
+pub enum TransmitError {
+    /// The transmit buffer is full, try again later.
+    BufferFull,
+}
+
+impl Display for TransmitError {
+    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
+        f.write_str(match self {
+            Self::BufferFull => "UART buffer full",
+        })
+    }
+}
+
+/// # MMIO version of an 8250 UART.
 ///
-/// **Noticed** This is only tested on the NS16550 compatible UART used in QEMU 5.0 virt machine of RISC-V
+/// **Note** This is only tested on the NS16550 compatible UART used in QEMU 5.0 virt machine of RISC-V.
 pub struct MmioUart8250<'a> {
     reg: &'a mut Registers,
 }
 
 impl<'a> MmioUart8250<'a> {
-    /// New a uart
-    pub fn new(base_address: usize) -> Self {
+    /// Creates a new UART.
+    ///
+    /// # Safety
+    ///
+    /// The given base address must point to the 8 MMIO control registers of an appropriate UART
+    /// device, which must be mapped into the address space of the process as device memory and not
+    /// have any other aliases.
+    pub unsafe fn new(base_address: usize) -> Self {
         Self {
-            reg: cast!(base_address),
+            reg: Registers::from_base_address(base_address),
         }
     }
 
-    /// A basic way to init the uart with interrupt enable
+    /// Initialises the UART with common settings and interrupts enabled.
     ///
-    /// Other way to init can be done by using other methods below
+    /// More customised initialisation can be done using other methods below.
     pub fn init(&self, clock: usize, baud_rate: usize) {
-        // Enable DLAB and Set divisor
-        self.enable_divisor_latch_accessible();
-        let divisor = clock / (16 * baud_rate);
-        self.write_dll(divisor as u8);
-        self.write_dlh((divisor >> 8) as u8);
+        // Enable DLAB and set divisor
+        self.set_divisor(clock, baud_rate);
 
         // Disable DLAB and set word length 8 bits, no parity, 1 stop bit
         self.write_lcr(3);
@@ -74,14 +149,20 @@ impl<'a> MmioUart8250<'a> {
         // self.enable_transmitter_holding_register_empty_interrupt();
     }
 
-    /// Set a new base_address
-    pub fn set_base_address(&mut self, base_address: usize) {
-        self.reg = cast!(base_address);
+    /// Sets a new base address for the UART.
+    ///
+    /// # Safety
+    ///
+    /// The given base address must point to the 8 MMIO control registers of an appropriate UART
+    /// device, which must be mapped into the address space of the process as device memory and not
+    /// have any other aliases.
+    pub unsafe fn set_base_address(&mut self, base_address: usize) {
+        self.reg = Registers::from_base_address(base_address);
     }
 
-    /// Read a byte from uart
+    /// Reads a byte from the UART.
     ///
-    /// Return `None` when data is not ready (RBR\[0\] != 1)
+    /// Returns `None` when data is not ready (RBR\[0\] != 1)
     pub fn read_byte(&self) -> Option<u8> {
         if self.is_data_ready() {
             Some(self.read_rbr())
@@ -90,11 +171,14 @@ impl<'a> MmioUart8250<'a> {
         }
     }
 
-    /// Write a byte to uart
-    ///
-    /// Error are not concerned now **MAYBE TODO**
-    pub fn write_byte(&self, byte: u8) {
-        self.write_thr(byte);
+    /// Writes a byte to the UART.
+    pub fn write_byte(&self, byte: u8) -> Result<(), TransmitError> {
+        if self.is_transmitter_holding_register_empty() {
+            self.write_thr(byte);
+            Ok(())
+        } else {
+            Err(TransmitError::BufferFull)
+        }
     }
 
     /// write THR (offset + 0)
@@ -108,20 +192,37 @@ impl<'a> MmioUart8250<'a> {
     /// > These registers really are the "heart" of serial data communication, and how data is transferred from your software to another computer and how it gets data from other devices. Reading and Writing to these registers is simply a matter of accessing the Port I/O address for the respective UART.
     /// >
     /// > If the receive buffer is occupied or the FIFO is full, the incoming data is discarded and the Receiver Line Status interrupt is written to the IIR register. The Overrun Error bit is also set in the Line Status Register.
-    pub fn write_thr(&self, value: u8) {
-        unsafe { self.reg.rw[0].write(value) }
+    #[inline]
+    fn write_thr(&self, value: u8) {
+        unsafe { self.reg.thr_rbr_dll.write(value) }
     }
 
     /// read RBR (offset + 0)
     ///
     /// Read Receiver Buffer to get data
-    pub fn read_rbr(&self) -> u8 {
-        self.reg.rw[0].read()
+    #[inline]
+    fn read_rbr(&self) -> u8 {
+        self.reg.thr_rbr_dll.read()
     }
 
-    /// read DLL (offset + 0)
+    /// write DLL (offset + 0)
     ///
-    /// get divisor latch low byte in the register
+    /// set divisor latch low byte in the register
+    #[inline]
+    fn write_dll(&self, value: u8) {
+        unsafe { self.reg.thr_rbr_dll.write(value) }
+    }
+
+    /// write DLH (offset + 1)
+    ///
+    /// set divisor latch high byte in the register
+    #[inline]
+    fn write_dlh(&self, value: u8) {
+        unsafe { self.reg.ier_dlh.write(value) }
+    }
+
+    /// Sets DLAB to true, sets divisor latch according to clock and baud_rate, then sets DLAB to
+    /// false.
     ///
     /// > ## Divisor Latch Bytes
     /// >
@@ -148,32 +249,7 @@ impl<'a> MmioUart8250<'a> {
     /// | 38400     | 3                    | $00                     | $03                    |
     /// | 57600     | 2                    | $00                     | $02                    |
     /// | 115200    | 1                    | $00                     | $01                    |
-    pub fn read_dll(&self) -> u8 {
-        self.reg.rw[0].read()
-    }
-
-    /// write DLL (offset + 0)
-    ///
-    /// set divisor latch low byte in the register
-    pub fn write_dll(&self, value: u8) {
-        unsafe { self.reg.rw[0].write(value) }
-    }
-
-    /// read DLH (offset + 1)
-    ///
-    /// get divisor latch high byte in the register
-    pub fn read_dlh(&self) -> u8 {
-        self.reg.rw[1].read()
-    }
-
-    /// write DLH (offset + 1)
-    ///
-    /// set divisor latch high byte in the register
-    pub fn write_dlh(&self, value: u8) {
-        unsafe { self.reg.rw[1].write(value) }
-    }
-
-    /// Set divisor latch according to clock and baud_rate, then set DLAB to false
+    #[inline]
     pub fn set_divisor(&self, clock: usize, baud_rate: usize) {
         self.enable_divisor_latch_accessible();
         let divisor = clock / (16 * baud_rate);
@@ -202,176 +278,124 @@ impl<'a> MmioUart8250<'a> {
     /// > | 2   | Enable Receiver Line Status Interrupt               |
     /// > | 1   | Enable Transmitter Holding Register Empty Interrupt |
     /// > | 0   | Enable Received Data Available Interrupt            |
-    pub fn read_ier(&self) -> u8 {
-        self.reg.rw[1].read()
+    #[inline]
+    fn read_ier(&self) -> u8 {
+        self.reg.ier_dlh.read()
     }
 
     /// Write IER (offset + 1)
     ///
     /// Write Interrupt Enable Register to turn on/off interrupts
+    #[inline]
     pub fn write_ier(&self, value: u8) {
-        unsafe { self.reg.rw[1].write(value) }
+        unsafe { self.reg.ier_dlh.write(value) }
+    }
+
+    /// Get IER bitflags
+    #[inline]
+    fn ier(&self) -> IER {
+        IER::from_bits_truncate(self.read_ier())
+    }
+
+    /// Set IER via bitflags
+    #[inline]
+    fn set_ier(&self, flag: IER) {
+        self.write_ier(flag.bits())
     }
 
     /// get whether low power mode (16750) is enabled (IER\[5\])
     pub fn is_low_power_mode_enabled(&self) -> bool {
-        self.reg.rw[1].read() & 0b0010_0000 != 0
-    }
-
-    /// toggle low power mode (16750) (IER\[5\])
-    pub fn toggle_low_power_mode(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v ^ 0b0010_0000) }
+        self.ier().contains(IER::LPM)
     }
 
     /// enable low power mode (16750) (IER\[5\])
     pub fn enable_low_power_mode(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v | 0b0010_0000) }
+        self.set_ier(self.ier() | IER::LPM)
     }
 
     /// disable low power mode (16750) (IER\[5\])
     pub fn disable_low_power_mode(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v & !0b0010_0000) }
+        self.set_ier(self.ier() & !IER::LPM)
     }
 
     /// get whether sleep mode (16750) is enabled (IER\[4\])
     pub fn is_sleep_mode_enabled(&self) -> bool {
-        self.reg.rw[1].read() & 0b0001_0000 != 0
-    }
-
-    /// toggle sleep mode (16750) (IER\[4\])
-    pub fn toggle_sleep_mode(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v ^ 0b0001_0000) }
+        self.ier().contains(IER::SM)
     }
 
     /// enable sleep mode (16750) (IER\[4\])
     pub fn enable_sleep_mode(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v | 0b0001_0000) }
+        self.set_ier(self.ier() | IER::SM)
     }
 
     /// disable sleep mode (16750) (IER\[4\])
     pub fn disable_sleep_mode(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v & !0b0001_0000) }
+        self.set_ier(self.ier() & !IER::SM)
     }
 
     /// get whether modem status interrupt is enabled (IER\[3\])
     pub fn is_modem_status_interrupt_enabled(&self) -> bool {
-        self.reg.rw[1].read() & 0b0000_1000 != 0
-    }
-
-    /// toggle modem status interrupt (IER\[3\])
-    pub fn toggle_modem_status_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v ^ 0b0000_1000) }
+        self.ier().contains(IER::MSI)
     }
 
     /// enable modem status interrupt (IER\[3\])
     pub fn enable_modem_status_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v | 0b0000_1000) }
+        self.set_ier(self.ier() | IER::MSI)
     }
 
     /// disable modem status interrupt (IER\[3\])
     pub fn disable_modem_status_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v & !0b0000_1000) }
+        self.set_ier(self.ier() & !IER::MSI)
     }
 
     /// get whether receiver line status interrupt is enabled (IER\[2\])
     pub fn is_receiver_line_status_interrupt_enabled(&self) -> bool {
-        self.reg.rw[1].read() & 0b0000_0100 != 0
-    }
-
-    /// toggle receiver line status interrupt (IER\[2\])
-    pub fn toggle_receiver_line_status_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v ^ 0b0000_0100) }
+        self.ier().contains(IER::RLSI)
     }
 
     /// enable receiver line status interrupt (IER\[2\])
     pub fn enable_receiver_line_status_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v | 0b0000_0100) }
+        self.set_ier(self.ier() | IER::RLSI)
     }
 
     /// disable receiver line status interrupt (IER\[2\])
     pub fn disable_receiver_line_status_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v & !0b0000_0100) }
+        self.set_ier(self.ier() & !IER::RLSI)
     }
 
     /// get whether transmitter holding register empty interrupt is enabled (IER\[1\])
     pub fn is_transmitter_holding_register_empty_interrupt_enabled(&self) -> bool {
-        self.reg.rw[1].read() & 0b0000_0010 != 0
-    }
-
-    /// toggle transmitter holding register empty interrupt (IER\[1\])
-    pub fn toggle_transmitter_holding_register_empty_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v ^ 0b0000_0010) }
+        self.ier().contains(IER::THREI)
     }
 
     /// enable transmitter holding register empty interrupt (IER\[1\])
     pub fn enable_transmitter_holding_register_empty_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v | 0b0000_0010) }
+        self.set_ier(self.ier() | IER::THREI)
     }
 
     /// disable transmitter holding register empty interrupt (IER\[1\])
     pub fn disable_transmitter_holding_register_empty_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v & !0b0000_0010) }
+        self.set_ier(self.ier() & !IER::THREI)
     }
 
     /// get whether received data available is enabled (IER\[0\])
     pub fn is_received_data_available_interrupt_enabled(&self) -> bool {
-        self.reg.rw[1].read() & 0b0000_0001 != 0
-    }
-
-    /// toggle received data available (IER\[0\])
-    pub fn toggle_received_data_available_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v ^ 0b0000_0001) }
+        self.ier().contains(IER::RDAI)
     }
 
     /// enable received data available (IER\[0\])
     pub fn enable_received_data_available_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v | 0b0000_0001) }
+        self.set_ier(self.ier() | IER::RDAI)
     }
 
     /// disable received data available (IER\[0\])
     pub fn disable_received_data_available_interrupt(&self) {
-        unsafe { self.reg.rw[1].modify(|v| v & !0b0000_0001) }
-    }
-
-    /// Read IIR (offset + 2)
-    ///
-    /// > ## Interrupt Identification Register
-    /// >
-    /// > Offset: +2 . This register is to be used to help identify what the unique characteristics of the UART chip that you are using has. This chip has two uses:
-    /// >
-    /// > - Identification of why the UART triggered an interrupt.
-    /// > - Identification of the UART chip itself.
-    /// >
-    /// > Of these, identification of why the interrupt service routine has been invoked is perhaps the most important.
-    /// >
-    /// > The following table explains some of the details of this register, and what each bit on it represents:
-    /// >
-    /// > | Bit        | Notes                             |       |                                   |                                              |                                                                                           |
-    /// > | ---------- | --------------------------------- | ----- | --------------------------------- | -------------------------------------------- | ----------------------------------------------------------------------------------------- |
-    /// > | 7 and 6    | Bit 7                             | Bit 6 |                                   |                                              |                                                                                           |
-    /// > |            | 0                                 | 0     | No FIFO on chip                   |                                              |                                                                                           |
-    /// > |            | 0                                 | 1     | Reserved condition                |                                              |                                                                                           |
-    /// > |            | 1                                 | 0     | FIFO enabled, but not functioning |                                              |                                                                                           |
-    /// > |            | 1                                 | 1     | FIFO enabled                      |                                              |                                                                                           |
-    /// > | 5          | 64 Byte FIFO Enabled (16750 only) |       |                                   |                                              |                                                                                           |
-    /// > | 4          | Reserved                          |       |                                   |                                              |                                                                                           |
-    /// > | 3, 2 and 1 | Bit 3                             | Bit 2 | Bit 1                             |                                              | Reset Method                                                                              |
-    /// > |            | 0                                 | 0     | 0                                 | Modem Status Interrupt                       | Reading Modem Status Register(MSR)                                                        |
-    /// > |            | 0                                 | 0     | 1                                 | Transmitter Holding Register Empty Interrupt | Reading Interrupt Identification Register(IIR) or Writing to Transmit Holding Buffer(THR) |
-    /// > |            | 0                                 | 1     | 0                                 | Received Data Available Interrupt            | Reading Receive Buffer Register(RBR)                                                      |
-    /// > |            | 0                                 | 1     | 1                                 | Receiver Line Status Interrupt               | Reading Line Status Register(LSR)                                                         |
-    /// > |            | 1                                 | 0     | 0                                 | Reserved                                     | N/A                                                                                       |
-    /// > |            | 1                                 | 0     | 1                                 | Reserved                                     | N/A                                                                                       |
-    /// > |            | 1                                 | 1     | 0                                 | Time-out Interrupt Pending (16550 & later)   | Reading Receive Buffer Register(RBR)                                                      |
-    /// > |            | 1                                 | 1     | 1                                 | Reserved                                     | N/A                                                                                       |
-    /// > | 0          | Interrupt Pending Flag            |       |                                   |                                              |                                                                                           |
-    pub fn read_iir(&self) -> u8 {
-        self.reg.rw[2].read()
+        self.set_ier(self.ier() & !IER::RDAI)
     }
 
     /// Read IIR\[7:6\] to get FIFO status
     pub fn read_fifo_status(&self) -> ChipFifoInfo {
-        match self.reg.rw[2].read() & 0b1100_0000 {
+        match self.reg.iir_fcr.read() & 0b1100_0000 {
             0 => ChipFifoInfo::NoFifo,
             0b0100_0000 => ChipFifoInfo::Reserved,
             0b1000_0000 => ChipFifoInfo::EnabledNoFunction,
@@ -382,27 +406,27 @@ impl<'a> MmioUart8250<'a> {
 
     /// get whether 64 Byte fifo (16750 only) is enabled (IIR\[5\])
     pub fn is_64byte_fifo_enabled(&self) -> bool {
-        self.reg.rw[2].read() & 0b0010_0000 != 0
+        self.reg.iir_fcr.read() & 0b0010_0000 != 0
     }
 
     /// Read IIR\[3:1\] to get interrupt type
-    pub fn read_interrupt_type(&self) -> InterruptType {
-        match self.reg.rw[2].read() & 0b0000_1110 {
-            0b0000 => InterruptType::ModemStatus,
-            0b0010 => InterruptType::TransmitterHoldingRegisterEmpty,
-            0b0100 => InterruptType::ReceivedDataAvailable,
-            0b0110 => InterruptType::ReceiverLineStatus,
-            0b1100 => InterruptType::Timeout,
-            0b1000 | 0b1010 | 0b1110 => InterruptType::Reserved,
-            _ => panic!("Can't reached"),
+    pub fn read_interrupt_type(&self) -> Option<InterruptType> {
+        let iir = self.reg.iir_fcr.read() & 0b0000_1111;
+        if iir & 1 != 0 {
+            None
+        } else {
+            match iir {
+                0b0000 => Some(InterruptType::ModemStatus),
+                0b0010 => Some(InterruptType::TransmitterHoldingRegisterEmpty),
+                0b0100 => Some(InterruptType::ReceivedDataAvailable),
+                0b0110 => Some(InterruptType::ReceiverLineStatus),
+                0b1100 => Some(InterruptType::Timeout),
+                0b1000 | 0b1010 | 0b1110 => Some(InterruptType::Reserved),
+                _ => panic!("Can't reached"),
+            }
         }
     }
 
-    /// get whether interrupt is pending (IIR\[0\])
-    pub fn is_interrupt_pending(&self) -> bool {
-        self.reg.rw[2].read() & 1 == 0
-    }
-
     /// Write FCR (offset + 2) to control FIFO buffers
     ///
     /// > ## FIFO Control Register
@@ -424,72 +448,32 @@ impl<'a> MmioUart8250<'a> {
     /// > | 2     | Clear Transmit FIFO         |       |                                   |                         |
     /// > | 1     | Clear Receive FIFO          |       |                                   |                         |
     /// > | 0     | Enable FIFOs                |       |                                   |                         |
-    pub fn write_fcr(&self, value: u8) {
-        unsafe { self.reg.rw[2].write(value) }
-    }
-
-    /// Read LCR (offset + 3)
-    ///
-    /// Read Line Control Register to get the data protocol and DLAB
-    ///
-    /// > ## Line Control Register
-    /// >
-    /// > Offset: +3 . This register has two major purposes:
-    /// >
-    /// > - Setting the Divisor Latch Access Bit (DLAB), allowing you to set the values of the Divisor Latch Bytes.
-    /// > - Setting the bit patterns that will be used for both receiving and transmitting the serial data. In other words, the serial data protocol you will be using (8-1-None, 5-2-Even, etc.).
-    /// >
-    /// > | Bit      | Notes                    |                              |             |               |
-    /// > | -------- | ------------------------ | ---------------------------- | ----------- | ------------- |
-    /// > | 7        | Divisor Latch Access Bit |                              |             |               |
-    /// > | 6        | Set Break Enable         |                              |             |               |
-    /// > | 3, 4 & 5 | Bit 5                    | Bit 4                        | Bit 3       | Parity Select |
-    /// > |          | 0                        | 0                            | 0           | No Parity     |
-    /// > |          | 0                        | 0                            | 1           | Odd Parity    |
-    /// > |          | 0                        | 1                            | 1           | Even Parity   |
-    /// > |          | 1                        | 0                            | 1           | Mark          |
-    /// > |          | 1                        | 1                            | 1           | Space         |
-    /// > | 2        | 0                        | One Stop Bit                 |             |               |
-    /// > |          | 1                        | 1.5 Stop Bits or 2 Stop Bits |             |               |
-    /// > | 0 & 1    | Bit 1                    | Bit 0                        | Word Length |               |
-    /// > |          | 0                        | 0                            | 5 Bits      |               |
-    /// > |          | 0                        | 1                            | 6 Bits      |               |
-    /// > |          | 1                        | 0                            | 7 Bits      |               |
-    /// > |          | 1                        | 1                            | 8 Bits      |               |
-    pub fn read_lcr(&self) -> u8 {
-        self.reg.rw[3].read()
+    #[inline]
+    fn write_fcr(&self, value: u8) {
+        unsafe { self.reg.iir_fcr.write(value) }
     }
 
     /// Write LCR (offset + 3)
     ///
     /// Write Line Control Register to set DLAB and the serial data protocol
-    pub fn write_lcr(&self, value: u8) {
-        unsafe { self.reg.rw[3].write(value) }
-    }
-
-    /// get whether DLAB is enabled
-    pub fn is_divisor_latch_accessible(&self) -> bool {
-        self.reg.rw[3].read() & 0b1000_0000 != 0
-    }
-
-    /// toggle DLAB
-    pub fn toggle_divisor_latch_accessible(&self) {
-        unsafe { self.reg.rw[3].modify(|v| v ^ 0b1000_0000) }
+    #[inline]
+    fn write_lcr(&self, value: u8) {
+        unsafe { self.reg.lcr.write(value) }
     }
 
     /// enable DLAB
-    pub fn enable_divisor_latch_accessible(&self) {
-        unsafe { self.reg.rw[3].modify(|v| v | 0b1000_0000) }
+    fn enable_divisor_latch_accessible(&self) {
+        unsafe { self.reg.lcr.modify(|v| v | 0b1000_0000) }
     }
 
     /// disable DLAB
-    pub fn disable_divisor_latch_accessible(&self) {
-        unsafe { self.reg.rw[3].modify(|v| v & !0b1000_0000) }
+    fn disable_divisor_latch_accessible(&self) {
+        unsafe { self.reg.lcr.modify(|v| v & !0b1000_0000) }
     }
 
     /// get parity of used data protocol
     pub fn get_parity(&self) -> Parity {
-        match self.reg.rw[3].read() & 0b0011_1000 {
+        match self.reg.lcr.read() & 0b0011_1000 {
             0b0000_0000 => Parity::No,
             0b0000_1000 => Parity::Odd,
             0b0001_1000 => Parity::Even,
@@ -502,11 +486,11 @@ impl<'a> MmioUart8250<'a> {
     /// set parity
     pub fn set_parity(&self, parity: Parity) {
         match parity {
-            Parity::No => unsafe { self.reg.rw[3].modify(|v| (v & 0b1100_0111)) },
-            Parity::Odd => unsafe { self.reg.rw[3].modify(|v| (v & 0b1100_0111) | 0b0000_1000) },
-            Parity::Even => unsafe { self.reg.rw[3].modify(|v| (v & 0b1100_0111) | 0b0001_1000) },
-            Parity::Mark => unsafe { self.reg.rw[3].modify(|v| (v & 0b1100_0111) | 0b0010_1000) },
-            Parity::Space => unsafe { self.reg.rw[3].modify(|v| v | 0b0011_1000) },
+            Parity::No => unsafe { self.reg.lcr.modify(|v| (v & 0b1100_0111)) },
+            Parity::Odd => unsafe { self.reg.lcr.modify(|v| (v & 0b1100_0111) | 0b0000_1000) },
+            Parity::Even => unsafe { self.reg.lcr.modify(|v| (v & 0b1100_0111) | 0b0001_1000) },
+            Parity::Mark => unsafe { self.reg.lcr.modify(|v| (v & 0b1100_0111) | 0b0010_1000) },
+            Parity::Space => unsafe { self.reg.lcr.modify(|v| v | 0b0011_1000) },
         }
     }
 
@@ -514,59 +498,38 @@ impl<'a> MmioUart8250<'a> {
     ///
     /// Simply return a u8 to indicate 1 or 1.5/2 bits
     pub fn get_stop_bit(&self) -> u8 {
-        ((self.reg.rw[3].read() & 0b100) >> 2) + 1
+        ((self.reg.lcr.read() & 0b100) >> 2) + 1
     }
 
     /// set stop bit, only 1 and 2 can be used as `stop_bit`
     pub fn set_stop_bit(&self, stop_bit: u8) {
         match stop_bit {
-            1 => unsafe { self.reg.rw[3].modify(|v| v & 0b1111_1011) },
-            2 => unsafe { self.reg.rw[3].modify(|v| v | 0b0000_0100) },
+            1 => unsafe { self.reg.lcr.modify(|v| v & 0b1111_1011) },
+            2 => unsafe { self.reg.lcr.modify(|v| v | 0b0000_0100) },
             _ => panic!("Invalid stop bit"),
         }
     }
 
     /// get word length of used data protocol
     pub fn get_word_length(&self) -> u8 {
-        (self.reg.rw[3].read() & 0b11) + 5
+        (self.reg.lcr.read() & 0b11) + 5
     }
 
     /// set word length, only 5..=8 can be used as `length`
     pub fn set_word_length(&self, length: u8) {
         if (5..=8).contains(&length) {
-            unsafe { self.reg.rw[3].modify(|v| v | (length - 5)) }
+            unsafe { self.reg.lcr.modify(|v| v | (length - 5)) }
         } else {
             panic!("Invalid word length")
         }
     }
 
-    /// Read MCR (offset + 4)
-    ///
-    /// Read Modem Control Register to get how flow is controlled
-    ///
-    /// > ## Modem Control Register
-    /// >
-    /// > Offset: +4 . This register allows you to do "hardware" flow control, under software control. Or in a more practical manner, it allows direct manipulation of four different wires on the UART that you can set to any series of independent logical states, and be able to offer control of the modem. It should also be noted that most UARTs need Auxiliary Output 2 set to a logical "1" to enable interrupts.
-    /// >
-    /// > | Bit | Notes                            |
-    /// > | --- | -------------------------------- |
-    /// > | 7   | Reserved                         |
-    /// > | 6   | Reserved                         |
-    /// > | 5   | Autoflow Control Enabled (16750) |
-    /// > | 4   | Loopback Mode                    |
-    /// > | 3   | Auxiliary Output 2               |
-    /// > | 2   | Auxiliary Output 1               |
-    /// > | 1   | Request To Send                  |
-    /// > | 0   | Data Terminal Ready              |
-    pub fn read_mcr(&self) -> u8 {
-        self.reg.rw[4].read()
-    }
-
     /// Write MCR (offset + 4)
     ///
     /// Write Modem Control Register to control flow
-    pub fn write_mcr(&self, value: u8) {
-        unsafe { self.reg.rw[4].write(value) }
+    #[inline]
+    fn write_mcr(&self, value: u8) {
+        unsafe { self.reg.mcr.write(value) }
     }
 
     /// Read LSR (offset + 5)
@@ -585,43 +548,52 @@ impl<'a> MmioUart8250<'a> {
     /// > | 2   | Parity Error                       |
     /// > | 1   | Overrun Error                      |
     /// > | 0   | Data Ready                         |
-    pub fn read_lsr(&self) -> u8 {
-        self.reg.ro[0].read()
+    #[inline]
+    fn read_lsr(&self) -> u8 {
+        self.reg.lsr.read()
+    }
+
+    /// Get LSR bitflags
+    #[inline]
+    fn lsr(&self) -> LSR {
+        LSR::from_bits_truncate(self.read_lsr())
     }
 
     /// get whether there is an error in received FIFO
     pub fn is_received_fifo_error(&self) -> bool {
-        self.reg.ro[0].read() & 0b1000_0000 != 0
+        self.lsr().contains(LSR::RFE)
     }
 
-    /// get whether data holding registers are empty
+    /// Gets whether data holding registers are empty, i.e. the UART has finished transmitting all
+    /// the data it has been given.
     pub fn is_data_holding_registers_empty(&self) -> bool {
-        self.reg.ro[0].read() & 0b0100_0000 != 0
+        self.lsr().contains(LSR::DHRE)
     }
 
-    /// get whether transmitter holding register is empty
+    /// Gets whether transmitter holding register is empty, i.e. the UART is ready to be given more
+    /// data to transmit.
     pub fn is_transmitter_holding_register_empty(&self) -> bool {
-        self.reg.ro[0].read() & 0b0010_0000 != 0
+        self.lsr().contains(LSR::THRE)
     }
 
     pub fn is_break_interrupt(&self) -> bool {
-        self.reg.ro[0].read() & 0b0001_0000 != 0
+        self.lsr().contains(LSR::BI)
     }
 
     pub fn is_framing_error(&self) -> bool {
-        self.reg.ro[0].read() & 0b0000_1000 != 0
+        self.lsr().contains(LSR::FE)
     }
 
     pub fn is_parity_error(&self) -> bool {
-        self.reg.ro[0].read() & 0b0000_0100 != 0
+        self.lsr().contains(LSR::PE)
     }
 
     pub fn is_overrun_error(&self) -> bool {
-        self.reg.ro[0].read() & 0b0000_0010 != 0
+        self.lsr().contains(LSR::OE)
     }
 
     pub fn is_data_ready(&self) -> bool {
-        self.reg.ro[0].read() & 0b0000_0001 != 0
+        self.lsr().contains(LSR::DR)
     }
 
     /// Read MSR (offset + 6)
@@ -640,88 +612,60 @@ impl<'a> MmioUart8250<'a> {
     /// > | 2   | Trailing Edge Ring Indicator |
     /// > | 1   | Delta Data Set Ready         |
     /// > | 0   | Delta Clear To Send          |
-    pub fn read_msr(&self) -> u8 {
-        self.reg.ro[1].read()
+    #[inline]
+    fn read_msr(&self) -> u8 {
+        self.reg.msr.read()
+    }
+
+    /// Get MSR bitflags
+    #[inline]
+    fn msr(&self) -> MSR {
+        MSR::from_bits_truncate(self.read_msr())
     }
 
     pub fn is_carrier_detect(&self) -> bool {
-        self.reg.ro[1].read() & 0b1000_0000 != 0
+        self.msr().contains(MSR::CD)
     }
 
     pub fn is_ring_indicator(&self) -> bool {
-        self.reg.ro[1].read() & 0b0100_0000 != 0
+        self.msr().contains(MSR::RI)
     }
 
     pub fn is_data_set_ready(&self) -> bool {
-        self.reg.ro[1].read() & 0b0010_0000 != 0
+        self.msr().contains(MSR::DSR)
     }
 
     pub fn is_clear_to_send(&self) -> bool {
-        self.reg.ro[1].read() & 0b0001_0000 != 0
+        self.msr().contains(MSR::CTS)
     }
 
     pub fn is_delta_data_carrier_detect(&self) -> bool {
-        self.reg.ro[1].read() & 0b0000_1000 != 0
+        self.msr().contains(MSR::DDCD)
     }
 
     pub fn is_trailing_edge_ring_indicator(&self) -> bool {
-        self.reg.ro[1].read() & 0b0000_0100 != 0
+        self.msr().contains(MSR::TERI)
     }
 
     pub fn is_delta_data_set_ready(&self) -> bool {
-        self.reg.ro[1].read() & 0b0000_0010 != 0
+        self.msr().contains(MSR::DDSR)
     }
 
     pub fn is_delta_clear_to_send(&self) -> bool {
-        self.reg.ro[1].read() & 0b0000_0001 != 0
+        self.msr().contains(MSR::DCTS)
     }
 
+    #[inline]
     pub fn read_sr(&self) -> u8 {
         self.reg.scratch.read()
     }
 
+    #[inline]
     pub fn write_sr(&self, value: u8) {
         unsafe { self.reg.scratch.write(value) }
     }
 }
 
-/// ## embedded-hal::serial::Read
-///
-/// This is a very simple implementation, based on [rustsbi/rustsbi-qemu](https://github.com/rustsbi/rustsbi-qemu/blob/main/rustsbi-qemu/src/ns16550a.rs#L36-L52)
-#[cfg(feature = "embedded")]
-impl<'a> serial::Read<u8> for MmioUart8250<'a> {
-    type Error = Infallible;
-
-    fn try_read(&mut self) -> nb::Result<u8, Self::Error> {
-        if self.is_interrupt_pending() {
-            Ok(self.read_rbr())
-        } else {
-            Err(nb::Error::WouldBlock)
-        }
-    }
-}
-
-/// ## embedded-hal::serial::Write
-///
-/// This is a very simple implementation, based on [rustsbi/rustsbi-qemu](https://github.com/rustsbi/rustsbi-qemu/blob/main/rustsbi-qemu/src/ns16550a.rs#L54-L75)
-#[cfg(feature = "embedded")]
-impl<'a> serial::Write<u8> for MmioUart8250<'a> {
-    type Error = Infallible;
-
-    fn try_write(&mut self, word: u8) -> nb::Result<(), Self::Error> {
-        self.write_thr(word);
-        Ok(())
-    }
-
-    fn try_flush(&mut self) -> nb::Result<(), Self::Error> {
-        if self.is_interrupt_pending() {
-            Ok(())
-        } else {
-            Err(nb::Error::WouldBlock)
-        }
-    }
-}
-
 /// ## fmt::Write
 ///
 /// A simple implementation, may be changed in the future
@@ -729,8 +673,90 @@ impl<'a> serial::Write<u8> for MmioUart8250<'a> {
 impl<'a> fmt::Write for MmioUart8250<'a> {
     fn write_str(&mut self, s: &str) -> fmt::Result {
         for c in s.as_bytes() {
-            self.write_thr(*c);
+            // If buffer is full, keep retrying.
+            while self.write_byte(*c) == Err(TransmitError::BufferFull) {}
         }
         Ok(())
     }
 }
+
+#[cfg(feature = "embedded")]
+impl embedded_hal::serial::Read<u8> for MmioUart8250<'_> {
+    type Error = Infallible;
+
+    fn read(&mut self) -> nb::Result<u8, Self::Error> {
+        self.read_byte().ok_or(nb::Error::WouldBlock)
+    }
+}
+
+#[cfg(feature = "embedded")]
+impl embedded_hal::serial::Write<u8> for MmioUart8250<'_> {
+    type Error = Infallible;
+
+    fn write(&mut self, byte: u8) -> nb::Result<(), Self::Error> {
+        self.write_byte(byte).map_err(|e| match e {
+            TransmitError::BufferFull => nb::Error::WouldBlock,
+        })
+    }
+
+    fn flush(&mut self) -> nb::Result<(), Self::Error> {
+        if self.is_data_holding_registers_empty() {
+            Ok(())
+        } else {
+            Err(nb::Error::WouldBlock)
+        }
+    }
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    // These tests treat normal memory as device memory, which is not necessarily guaranteed to
+    // work, but it seems to for now.
+
+    #[test]
+    fn initialise() {
+        // Create a fake UART using an in-memory buffer, and check that it is initialised as
+        // expected.
+        let mut fake_registers: [u8; 8] = [0xff; 8];
+        let uart = unsafe { MmioUart8250::new(&mut fake_registers as *mut u8 as usize) };
+
+        uart.init(11_059_200, 115200);
+
+        assert!(matches!(uart.get_parity(), Parity::No));
+        assert_eq!(uart.get_stop_bit(), 1);
+        assert_eq!(uart.get_word_length(), 8);
+    }
+
+    #[test]
+    fn write() {
+        let mut fake_registers: [u8; 8] = [0; 8];
+        let uart = unsafe { MmioUart8250::new(&mut fake_registers as *mut u8 as usize) };
+
+        // Pretend that the transmit buffer is full.
+        fake_registers[5] = 0;
+        assert_eq!(uart.write_byte(0x42), Err(TransmitError::BufferFull));
+        assert_eq!(fake_registers[0], 0);
+
+        // Pretend that the transmit buffer is available.
+        fake_registers[5] = 0b0010_0000;
+        assert_eq!(uart.write_byte(0x42), Ok(()));
+        assert_eq!(fake_registers[0], 0x42);
+    }
+
+    #[test]
+    fn read() {
+        let mut fake_registers: [u8; 8] = [0; 8];
+        let uart = unsafe { MmioUart8250::new(&mut fake_registers as *mut u8 as usize) };
+
+        // First try to read when there is nothing available.
+        assert_eq!(uart.read_byte(), None);
+
+        // Set the UART up to have a byte available to read and read it.
+        fake_registers[0] = 0xab;
+        fake_registers[5] = 0b0000_0001;
+
+        assert_eq!(uart.read_byte(), Some(0xab));
+    }
+}
